//20/06/2000
//IMPLEMENTAZIONE DEL THREAD DI GESTIONE DELLE PRENOTAZIONI

import java.net.*;
import java.io.IOException;

public class DeamonBook extends Thread
	{
	private GroupRequest GRdo,GRundo;		//gruppo di richieste sospese di do e undo
	private DistribuitedDB mdb;			//riferimento al thread di modifica del db
	private Db database;				//database dei voli
	private DatagramSocket UDPSock;		//socket UDP per la comunicazione tra server
	private InetAddress SuccHost, PrecHost; 		//riferimenti agli indirizzi di successore e predecessore
	private ServerConfig HostConfig;		//riferimento al configuratore del server
		
	DeamonBook(ServerConfig c)
		{
		super("DEAMON BOOK");
		HostConfig=c;
		}
	//procedure di inserimento nelle code di sospensione e invio datagrammi (scrittura) sulle connessione UDP		
	/* sia le procedure di registrazione che quelle di invio saranno utilizzate dai DistribuitedDo */
	public synchronized boolean putRequestDo(RequestThreadDo r)		{return GRdo.putRq(r);}
	public synchronized boolean putRequestUndo(RequestThreadUndo r)	{return GRundo.putRq(r);}
	public synchronized void sendOnDirect(byte[] p) throws IOException
		{
		DatagramPacket dat=new DatagramPacket(p, p.length, SuccHost, CServer.PORTUDP);
		UDPSock.send(dat);
		}
	public synchronized void sendOnIndirect(byte[] p) throws IOException
		{
		DatagramPacket dat=new DatagramPacket(p, p.length, PrecHost, CServer.PORTUDP);
		UDPSock.send(dat);
		}
	public synchronized void putRequestDB(DistribuitedDB d)		{mdb=d;}
	//ridefinizione indirizzi di connessione con i vicini
	public synchronized void redefineSuccHost()				{SuccHost=HostConfig.getSuccHost();}
	public synchronized void redefinePrecHost()				{PrecHost=HostConfig.getPrecHost();}
	
	public void run()
		{
		Log.netLogging("@@@@@@ "+getName()+": Active!!");
		//riferimento al database
		database=HostConfig.getDatabase();
		//creazione code di sospensione per richieste distribuite
		GRdo= new GroupRequest();
		GRundo=new GroupRequest();
		try {
			UDPSock=new DatagramSocket(CServer.PORTUDP);
			Log.netLogging(getName()+": UDP connection created at "+CServer.PORTUDP);
			while (true)	//ciclo infinito
				{
				byte[] cont=new byte[CServer.MAX_DIM_PACK];
				DatagramPacket rpack=new DatagramPacket(cont, CServer.MAX_DIM_PACK);
				//ricezione messaggio (da qualsiasi server)
				UDPSock.receive(rpack);
				byte[] rdata=rpack.getData();
				if (rpack.getAddress().getHostAddress().equals(PrecHost.getHostAddress()))
					{//ricevuto dal predecessore - verso diretto
					//riconoscimento messaggio
					if  ((Message.decodeMsg(cont[0])).equals("DOB"))
						{//DO_BOOKING
						MsgDoBooking rmsg=MsgDoBooking.UnByting(rdata);
						Log.srvLogging(getName()+": receive DO_BOOKING from predecessor");
						if (rmsg.getIPaddrSender().equals(InetAddress.getLocalHost().getHostAddress()))
							{//messaggio originato da questo server
							int pos_do=GRdo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]) );
							if (pos_do>=0 )
								{//esiste tra le sospese di "do"
								//estrazione thread richiedente (DistribuitedDo)
								DistribuitedDo dd=((DistribuitedDo) ((RequestThreadDo) GRdo.getRq(pos_do)).getThread());
								//elimina dalla coda dei sospesi della richiesta
								GRdo.pushRq(pos_do);
								//settaggio risposta al ThreadService relativo a quel DistribuitedDo
								ThreadService ts=((ThreadService) dd.getService());
								ts.setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], true, "DO"));
								Log.srvLogging(ts.getName()+" resume with success (DO)");
								//terminazione DistribuitedDo
								dd.stop();
								}
							else	
								Log.srvLogging("($$$) "+getName()+": REQUEST NOT SUSPEND [do]"); //ERRORE
								//richiesta non sospesa
							}
						else 
							{//messaggio originato in un'altro server
							int pos_do=GRdo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]));
							if (pos_do>=0)
								{//esiste nelle sospese di "do"
								if (rmsg.getPriority()>HostConfig.getPriorityHost())
									{//prenotazione locale perdente
									ClientKey localCKey=new ClientKey();
									Log.srvLogging(getName()+": collision loose....");
									boolean exist=false;
									int pos_fly;
									synchronized(database)
										{//accesso esclusivo al database: consistenza
										pos_fly=database.posFlight(rmsg.getFlyId());
										if (pos_fly>=0)
											{
											//ottenimento chiave identificativo cliente locale prenotato
											localCKey=database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]).getClientKey();
											//sovrascrittura prenotazione ricevuta
											database.addBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1], rmsg.getClient());
											exist=true;
											}
										}
									if (exist)
										{
										Log.srvLogging(getName()+": override booking....");
										//creazione DO_BOOKING_OVERRIDE e invio
										MsgDoBookingOverride smsg=new MsgDoBookingOverride(rmsg);
										sendOnDirect(smsg.Byting());
										Log.srvLogging(getName()+": send DO_BOOKING_OVERRIDE at successor");
										//estrazione del thread richiedente
										DistribuitedDo dd=((DistribuitedDo) ((RequestThreadDo) GRdo.getRq(pos_do)).getThread()); 
										//elimina dalla coda dei sospesi della richiesta
										GRdo.pushRq(pos_do);
										//risposta al ThreadService
										ThreadService ts=((ThreadService) dd.getService());
										ts.setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], false, "DO"));
										Log.srvLogging(ts.getName()+": resume with fail (DO)");
										//terminazione DistribuitedDo
										dd.stop();
										//generazione gestore speciale di prenotazione perdente
										MsgUndoBooking smsg2=new MsgUndoBooking(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], localCKey, InetAddress.getLocalHost().getHostAddress());
										GestDoLoose tdl=new GestDoLoose(HostConfig, smsg2);
										tdl.start();
										}
									else 	
										Log.srvLogging("($$$) "+getName()+": FLY NOT EXIST [do]"); //ERRORE volo inesitente
									}
								else //prenotazione locale vincente
									Log.srvLogging(getName()+": collision win...."); 
									//distruggi il messaggio arrivato
								}
							else	
								{//non esiste tra i "do"
								int pos_undo=GRundo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]) );
								if (!(pos_undo>=0))
									{
									//non esiste tra le sospese di "undo"
									boolean exist=false;
									boolean free=false;
									int pos_fly;
									synchronized(database)
										{
										pos_fly=database.posFlight(rmsg.getFlyId());
										if (pos_fly>=0)
											{//esiste il volo
											exist=true;
											if (!(database.isBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1])))
												{//il posto  libero
												//scrittura prenotazione sul DB
												database.addBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1], rmsg.getClient());
												free=true;
												}	
											}
										}
									if (exist)
										if (free)
											{//prenotazione locale effettuata
											//forword
											Log.srvLogging(getName()+": local booking OK");
											sendOnDirect(rmsg.Byting());
											Log.srvLogging(getName()+": forword DO_BOOKING at successor");
											}
										else 	
											{//posto occupato
											ClientAttrib c=database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]);
											if (c.Equals(rmsg.getClient()))
												{//prenotazione gi esistente
												//forword
												sendOnDirect(rmsg.Byting());
												Log.srvLogging(getName()+": forword DO_BOOKING at successor");
												}
											else //altro cliente che lo occupa
												//distruggi il messaggio
												Log.srvLogging("($$$) "+getName()+": ANOTHER BOOK: wait retrasmission");
											}
									else //volo inesistente 
										Log.srvLogging("($$$) "+getName()+": FLY NOT EXIST [do]"); 
									}
								else //esiste tra gli "undo"
									//distruggi il messaggio
									Log.srvLogging("($$$) "+getName()+": UNDO BOOKING [do]");
									
								}	
							}				
						}
					if  ((Message.decodeMsg(cont[0])).equals("DBO"))
						{
						MsgDoBookingOverride rmsg=MsgDoBookingOverride.UnByting(rdata);
						Log.srvLogging(getName()+": receive DO_BOOKING_OVERRIDE from predecessor");
						if (rmsg.getIPaddrSender().equals(InetAddress.getLocalHost().getHostAddress()))
							{//messaggio generato da questo server
							int pos_do=GRdo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]));
							if (pos_do>=0)
								{//esiste tra i sospesi di "do"
								//estrazione thread richiedente
								DistribuitedDo dd=((DistribuitedDo) ((RequestThreadDo) GRdo.getRq(pos_do)).getThread());
								//cancellazione richiesta
								GRdo.pushRq(pos_do);
								ThreadService ts=((ThreadService) dd.getService()); 
								ts.setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], true, "DO"));
								Log.srvLogging(ts.getName()+": resume with success (DO)");
								dd.stop();
								}
							else
								Log.srvLogging("($$$) "+getName()+": REQUEST NOT SUSPEND [do-over]"); //ERRORE non sospesa
							}
						else
							{//atro server generatore
							ClientKey localCKey=new ClientKey();;
							boolean over=false;
							synchronized(database)	{
								int pos_fly=database.posFlight(rmsg.getFlyId());
								if (pos_fly>=0)
									{//esiste il volo
									//ottenimento chiave identificativo cliente locale prenotato
									localCKey=database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]).getClientKey();
									//sovrascrittura
									database.addBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1], rmsg.getClient());
									over=true;
									}
										}
							if (over)
								{//sovrascrittura effettuata
								Log.srvLogging(getName()+": local booking OVERRIDE");
								int pos_do=GRdo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]));
								if (pos_do>=0)
									{//esite tra i "do"
									//estrazione del thread richiedente
									DistribuitedDo dd=( (DistribuitedDo) ( (RequestThreadUndo) GRdo.getRq(pos_do) ).getThread() ); 
									//elimina dalla coda dei sospesi della richiesta
									GRdo.pushRq(pos_do);
									//risveglio thread con risposta
									ThreadService ts=((ThreadService) dd.getService()); 
									ts.setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], false, "DO"));
									Log.srvLogging(ts.getName()+": resume with fail (DO)");
									dd.stop();
									//generazione processo speciale di gestione prenotazione perdente
									MsgUndoBooking smsg=new MsgUndoBooking(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], localCKey, InetAddress.getLocalHost().getHostAddress());
									GestDoLoose tdl=new GestDoLoose(HostConfig, smsg);
									tdl.start();
									}
								//forword
								sendOnDirect(rmsg.Byting());
								Log.srvLogging(getName()+": forword DO_BOOKING_OVERRIDE at successor");
								}
							else	
								Log.srvLogging("($$$) "+getName()+": FLY NOT EXIST [do-over]"); //ERRORE volo inesistente
							}
						}
					if  ((Message.decodeMsg(cont[0])).equals("UDB"))
						{
						MsgUndoBooking rmsg=MsgUndoBooking.UnByting(rdata);
						if (SuccHost.getHostAddress().equals(PrecHost.getHostAddress()))
							{//solo 2 server!!!! - prec=succ
							Log.srvLogging(getName()+": receive UNDO_BOOKING from successor");
							if (rmsg.getIPaddrSender().equals(InetAddress.getLocalHost().getHostAddress()))
								{//generato da questo server
								int pos_undo=GRundo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]));
								if (pos_undo>=0)
									{//esiste tra i sospesi di "undo"
									//estrazione thread richiedente
									DistribuitedUndo du=((DistribuitedUndo) ((RequestThreadUndo) GRundo.getRq(pos_undo)).getThread());
									//cancellazione richiesta
									GRundo.pushRq(pos_undo);
									if ((du.getService()) instanceof ThreadService)
										((ThreadService) du.getService()).setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], true, "UNDO"));
									Log.srvLogging(du.getService().getName()+": resume with success (UNDO)");
									du.stop();
									}
								else	
									Log.srvLogging("($$$) "+getName()+": REQUEST NOT SUSPEND [undo]");
								}
							else 
								{//altro server generatore
								boolean exist=false;
								int pos_fly;
								synchronized(database)
									{
									pos_fly=database.posFlight(rmsg.getFlyId());
									if (pos_fly>=0)
										if (rmsg.getClientId().Equals(database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]).getClientKey()))
											{//esiste la prenotazione
											database.delBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]);
											exist=true;
											}
									}
								if (exist)
									{//forword
									sendOnIndirect(rmsg.Byting());
									Log.srvLogging(getName()+": forword UNDO_BOOKING at predecessor");
									}
								else	
									if (pos_fly<0)
										Log.srvLogging("($$$) "+getName()+": FLY NOT EXIST [undo]"); //ERRORE non esiste il volo
									else
										//non esiste la prenotazione (undo ritrasmesso)
										{//forword
										Log.srvLogging("?? "+getName()+": local booking not exist");
										sendOnIndirect(rmsg.Byting());
										Log.srvLogging(getName()+": forword UNDO_BOOKING at predecessor");
										}
								}
							}
						else 
							Log.srvLogging("($$$) "+getName()+": WRONG DIRECTION [undo]"); //ERRORE!!!
						}
					if  ((Message.decodeMsg(cont[0])).equals("IFY"))
						{
						MsgInsertFly rmsg=MsgInsertFly.UnByting(rdata);
						Log.srvLogging(getName()+": receive INSERT_FLY from predecessor");
						if (rmsg.getIPaddrInsert().equals(InetAddress.getLocalHost().getHostAddress()))
							//generato da questo server
							mdb.stop();
						else //generato da altro server
							{
							boolean boo=false;
							synchronized(database)
									{
									if (!(database.posFlight(rmsg.getFlyId())>=0))
										{//volo non gia' esistente
										database.addFlight(new FlightAttrib(rmsg.getFlyId(), rmsg.getArrive(), rmsg.getLeave(), rmsg.getStartDate()));
										Log.srvLogging(getName()+": insert fly in local database");
										}
									}
							sendOnDirect(rmsg.Byting());
							Log.srvLogging(getName()+": forword INSERT_FLY at successor");
							}
						}
					if  ((Message.decodeMsg(cont[0])).equals("DFY"))
						{
						MsgDeleteFly rmsg=MsgDeleteFly.UnByting(rdata);
						Log.srvLogging(getName()+": receive DELETE_FLY from predecessor");
						if (rmsg.getIPaddrDelete().equals(InetAddress.getLocalHost().getHostAddress()))
							//generato da questo server
							mdb.stop();
						else //generato da altro server
							{
							synchronized(database)
								{
								int p=database.posFlight(rmsg.getFlyId());
								if (p>=0)
									{
									//volo esistente
									database.delFlight(p);
									Log.srvLogging(getName()+": delete fly from local database");
									}
								}
							sendOnDirect(rmsg.Byting());
							Log.srvLogging(getName()+": forword DELET_FLY at successor");
							}
						}
					}
				else	
					{//ricevuto non dal predecessore
					if (rpack.getAddress().getHostAddress().equals(SuccHost.getHostAddress()))
						{//ricevuto dal successore
						if  ((Message.decodeMsg(cont[0])).equals("DOB"))
							{//DO_BOOKING
							MsgDoBooking rmsg=MsgDoBooking.UnByting(rdata);
							Log.srvLogging("($$$) "+getName()+": WRONG DIRECTION [do]");
							}
						if ((Message.decodeMsg(cont[0])).equals("DBO"))
							{
							MsgDoBookingOverride rmsg=MsgDoBookingOverride.UnByting(rdata);
							Log.srvLogging("($$$) "+getName()+": WRONG DIRECTION [do-over]");
							}
						if  ((Message.decodeMsg(cont[0])).equals("IFY"))
							{
							MsgInsertFly rmsg=MsgInsertFly.UnByting(rdata);
							Log.srvLogging("($$$) "+getName()+": WRONG DIRECTION [ins]");
							}
						if  ((Message.decodeMsg(cont[0])).equals("DFY"))
							{
							MsgDeleteFly rmsg=MsgDeleteFly.UnByting(rdata);
							Log.srvLogging("($$$) "+getName()+": WRONG DIRECTION [del]");
							}
						if ((Message.decodeMsg(cont[0])).equals("UDB"))
							{
							//messaggio di UNDO_BOOOKING
							MsgUndoBooking rmsg=MsgUndoBooking.UnByting(rdata);
							Log.srvLogging(getName()+": receive UNDO_BOOKING from successor");
							if (rmsg.getIPaddrSender().equals(InetAddress.getLocalHost().getHostAddress()))
								{//generato da questo server
								int pos_undo=GRundo.posRq(new Request(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1]));
								if (pos_undo>=0)
									{//esiste tra i sospesi di "undo"
									//estrazione thread richiedente
									DistribuitedUndo du=((DistribuitedUndo) ((RequestThreadUndo) GRundo.getRq(pos_undo)).getThread());
									//cancellazione richiesta
									GRundo.pushRq(pos_undo);
									if ((du.getService()) instanceof ThreadService)
										((ThreadService) du.getService()).setResponse(new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], true, "UNDO"));
									Log.srvLogging(du.getService().getName()+" resume with success (UNDO)");
									du.stop();
									}
								else	
									Log.srvLogging("($$$) "+getName()+": REQUEST NOT SUSPEND [undo]"); //ERRORE
								}
							else 
								{
								boolean exist=false;//esistenza prenotazione
								//altro server generatore
								int pos_fly;
								synchronized(database)	
									{
									pos_fly=database.posFlight(rmsg.getFlyId());
									if (pos_fly>=0)
										if (rmsg.getClientId().Equals(database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]).getClientKey()))
											{//esiste la prenotazione
											database.delBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]);
											exist=true;
											}
									}
								if (exist)
									{//forword
									Log.srvLogging(getName()+": local booking undo");
									sendOnIndirect(rmsg.Byting());
									Log.srvLogging(getName()+": forword UNDO_BOOKING at predecessor");
									}
								else	
									if (pos_fly<0)
										Log.srvLogging("($$$) "+getName()+": FLY NOT EXIST [undo]"); //ERRORE non esiste il volo
									else
										//non esiste la prenotazione (undo ritrasmesso)
										{//forword
										Log.srvLogging("?? "+getName()+": local booking not exist");
										sendOnIndirect(rmsg.Byting());
										Log.srvLogging(getName()+": forword UNDO_BOOKING at predecessor");
										}
										
								}
							}
						}
					else	
						Log.srvLogging("($$$) "+getName()+": NOT SYSTEM-SERVER SENDER");//ERRORE ne successore ne predecessore come mittente
					}		
				}
			}//end try
		catch (SocketException e) 		{Log.netLogging("($$$) - [CREATE DATAGRAM SOCKET] "+getName()+": "+e);}
		catch (UnknownHostException e) 		{Log.netLogging("($$$) "+getName()+": "+e);}
		catch (IOException e)			{Log.netLogging("($$$) - [RECEIVE EXCEPTION] "+getName()+": "+e);}
		catch (Exception e) 			{Log.netLogging("($$$) "+getName()+": "+e);}
		}
	}