//16/06/2000
// IMPLEMENTAZIONE DEL THREAD DI SERVIZIO DELLE RICHIESTE DEL CLIENT
import java.io.*;
import java.net.*;

public class ThreadService extends Thread
	{
	private Socket ClientSock;		//riferimento alla socket del cliente generata dalla accept del listener
	private ObjectOutputStream ClientWR;	//stream di scrittura sul client
	private ObjectInputStream ClientRD;	//stream di lettura dal client
	private ServerConfig HostConfig;		//riferimento alla configurazione del server
	
	private MsgResponse Rsp;		//messaggio di risposta all'utente dell'ultima azione comandata
	private Db database;			//riferimento al database locale dei voli
	private DeamonBook forworder;		//riferimento al demone delle prenotazioni (do e undo)
	
	ThreadService(Socket s, ServerConfig c)
		{
		super();
		ClientSock=s;
		setName(s.getInetAddress().getHostName()+"-"+s.getPort());
		HostConfig=c;
		forworder=c.getDeamonBook();
		database=c.getDatabase();
		setName("Thread-SERVICE["+getName()+"]");
		}
	
	public synchronized void setResponse(MsgResponse s)	{Rsp=s;}
	//settaggio della risposta all'utente per prenotazioni e cancellazioni

	
	public void run()
		{
		try {
		ClientSock.setSoTimeout(CServer.TIMEOUT_SERVER);
		//creazione socket TCP stream connesse con il client: apertura in lettura e scrittura (visibili solo da questo processo)
		ClientWR=new ObjectOutputStream(ClientSock.getOutputStream());
		ClientWR.flush();
		ClientRD=new ObjectInputStream(ClientSock.getInputStream());
		
		Log.srvLogging(getName()+": Service for client started");
		Log.srvLogging("%%  "+ClientSock.toString());
		while(true)		//ciclo infinito
			{
			boolean okmsgrcv=false;	//messaggio ricevuto consistente con il protocollo
			//lettura byte di riconoscimento del messaggio
			byte type=ClientRD.readByte();
			//identificazione messaggio arrivato
			String typemsg=Message.decodeMsg(type);

			if (typemsg.equals("GFL"))	 
				{//GET_FLY_LIST
				try {	//ricostruzione messaggio arrivato
					okmsgrcv=true;
					MsgGetFlyList rmsg=(MsgGetFlyList) ClientRD.readObject();
					Log.srvLogging(getName()+": request fly list");
					//creazione messaggio di risposta FLY_LIST
					MsgFlyList smsg=new MsgFlyList(database.getFlightList());
					//scrittura sulla socket connesssa con il client del messaggio creato
					//non serve sincronizzazione in quanto  l'unico processo lettore e scrittore da questo end-point
					ClientWR.writeByte(Message.codeMsg("FL"));
					ClientWR.writeObject(smsg);
					Log.srvLogging(getName()+": send fly list");
					}
				catch (ClassNotFoundException e) {Log.srvLogging("($$$) - [FLY-LIST] "+getName()+" "+e);}
				}
				
			if (typemsg.equals("GFS"))
				{//GET_FREE_SEAT
				try {
					okmsgrcv=true;
					//ricostruzione messaggio arrivato
					MsgGetFreeSeat rmsg=(MsgGetFreeSeat) ClientRD.readObject();
					Log.srvLogging(getName()+": request seat free list for fly "+rmsg.getFlyId().getKey());
					//creazione messaggio di risposta SEAT_LIST
					MsgSeatList smsg;
					synchronized(database)
					//sincronizzazione per avere posizione consistente
						{
						int _pf=database.posFlight(rmsg.getFlyId());
						smsg=new MsgSeatList(rmsg.getFlyId(), database.getFreeSeatList(_pf), database.getFlight(_pf).getNumRow(), database.getFlight(_pf).getNumNum());
						}
					//scrittura sulla socket connessa con il client del messaggio creato
					ClientWR.writeByte(Message.codeMsg("SL"));
					ClientWR.writeObject(smsg);
					Log.srvLogging(getName()+": send free seat list");
					}
				catch (ClassNotFoundException e) 	{Log.srvLogging("($$$) - [GET_FREE_SEAT] "+getName()+" "+e);}
				catch (Exception e) 			{Log.srvLogging("($$$) - [GET_FREE_SEAT] "+getName()+" "+e);}
				}
				
			if (typemsg.equals("BT"))	
				{//BOOKING THIS
				try {
					okmsgrcv=true;
					//ricostruzione messaggio arrivato
					MsgBookingThis rmsg=(MsgBookingThis) ClientRD.readObject();
					Log.srvLogging(getName()+": request for booking - fly: "+rmsg.getFlyId().getKey()+" ["+rmsg.getSeat()[0]+" & "+rmsg.getSeat()[1]+"]");
					boolean exist=false;
					boolean book=false;
					synchronized(database)
					//sincronizzazione per ottenere informazioni consistenti di posizione, prenotazione ecc.
						{
						int pos_fly=database.posFlight(rmsg.getFlyId());
						if (pos_fly>=0)
							{//volo esistente
							exist=true;
							if (!(database.isBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1])))
								{//posto libero
								book=true;
								//occupazione locale del posto
								database.addBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1], rmsg.getClient());
								Log.srvLogging(getName()+": local booking OK");
								}
							}
						}
					if (exist)
						{//volo esistente
						if (book)
							{//prenotazione locale e` stata effetuata
							boolean goon=false;
							MsgDoBooking smsg=new MsgDoBooking(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], rmsg.getClient(), InetAddress.getLocalHost().getHostAddress(),  HostConfig.getPriorityHost());
							//generazione processo di richiesta prenotazione distribuita
							DistribuitedDo regt=new DistribuitedDo(smsg, this, HostConfig);
							Log.srvLogging(getName()+": request distribuited booking");
							Log.srvLogging(getName()+": .......");
							regt.start();
							try {regt.join();}
							catch (InterruptedException e) {/*ignored*/}
							goon=true;
							//NB. la risposta (Rsp) e settata dal DeamonBook  o dal DistribuitedDo
							if (!Rsp.isSuccesseful())
								//insuccesso della prenotazione
								synchronized(database)
									{//sincronizzazione per la cancellazione locale
									int pos_fly=database.posFlight(rmsg.getFlyId());
									if (pos_fly>=0)
										{
										database.delBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]);
										Log.srvLogging(getName()+": booking FAIL");
										}
									}
							else
								Log.srvLogging(getName()+": booking OK");
							}
						else	//prenotazione locale non effettuata (concorrenza locale)
							{
							Rsp=new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], false,"DO");
							Log.srvLogging(getName()+": booking FAIL");
							}
						//invio al cliente della risposta
						ClientWR.writeByte(Message.codeMsg("RSP"));
						ClientWR.writeObject(Rsp);
						Log.srvLogging(getName()+": send response");
						}
					else	
						Log.srvLogging("($$$) - [BOOKING_THIS]"+getName()+":  FLY NOT EXIST");	
					}
				catch (InterruptedException e) 		{Log.srvLogging("($$$) - [BOOKING_THIS] "+getName()+" "+e);}
				catch (ClassNotFoundException e) 	{Log.srvLogging("($$$) - [BOOKING_THIS] "+getName()+" "+e);}
				catch (Exception e) 			{Log.srvLogging("($$$) - [BOOKING_THIS] "+getName()+" "+e);}
				}
				
			if (typemsg.equals("GCFL"))
				{//GET_CLIENT_FLY_LIST
				try {
					okmsgrcv=true;
					//ricostruzione messaggio arrivato
					MsgGetClientFlyList rmsg=(MsgGetClientFlyList) ClientRD.readObject();
					Log.srvLogging(getName()+": request for booking fly list");
					//creazione messaggio di risposta FLY_LIST
					MsgFlyList smsg=new MsgFlyList(database.getClientFly(rmsg.getClientId()));
					//scrittura sulla socket connesssa con il client del messaggio creato
					ClientWR.writeByte(Message.codeMsg("FL"));
					ClientWR.writeObject(smsg);
					Log.srvLogging(getName()+": send booking fly list");
					}
				catch (ClassNotFoundException e) {Log.srvLogging("($$$) - [GET_CLIENT_FLY_LIST] "+getName()+": "+e);}
				}
				
			if (typemsg.equals("GCSL"))	
				{//GET_CLIENT_SEAT_LIST
				try {
					okmsgrcv=true;
					//ricostruzione messaggio arrivato
					MsgGetClientSeatList rmsg=(MsgGetClientSeatList) ClientRD.readObject();
					Log.srvLogging(getName()+": request for client seat list for fly "+rmsg.getFlyId().getKey());
					//creazione messaggio di risposta SEAT_LIST
					MsgSeatList smsg;
					synchronized(database)
						{
						int _pf=database.posFlight(rmsg.getFlyId());
						smsg=new MsgSeatList(rmsg.getFlyId(), database.getClientSeat(_pf, rmsg.getClientId()), database.getFlight(_pf).getNumRow(), database.getFlight(_pf).getNumNum());
						}
					//scrittura sulla socket connesssa con il client del messaggio creato
					ClientWR.writeByte(Message.codeMsg("SL"));
					ClientWR.writeObject(smsg);
					Log.srvLogging(getName()+": send client seat list");
					}
				catch (ClassNotFoundException e)	{Log.srvLogging("($$$) - [GET_CLIENT_SEAT_LIST] "+getName()+": "+e);}
				catch (Exception e) 			{Log.srvLogging("($$$) - [GET_CLIENT_SEAT_LIST] "+getName()+" "+e);}
				}
				
			if (typemsg.equals("DB"))
				{//DELETE_BOOKING
				try {
					okmsgrcv=true;
					//ricostruzione messaggio arrivato
					MsgDeleteBooking rmsg=(MsgDeleteBooking) ClientRD.readObject();
					Log.srvLogging(getName()+": request for delete booking - fly: "+rmsg.getFlyId().getKey()+" ["+rmsg.getSeat()[0]+"."+rmsg.getSeat()[1]+"]");
					boolean exist=false;
					synchronized(database)
						{
						int pos_fly=database.posFlight(rmsg.getFlyId());
						if (pos_fly>=0)
							//volo esistente
							try {if (database.getClientAttrib(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]).getClientKey().getKey()==rmsg.getClientId().getKey())
									exist=true;}
							catch (ArrayIndexOutOfBoundsException e) 
								{Rsp=new MsgResponse(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], false,  "Not Exist");}
						}
					if (exist)
						{
						boolean goon=false;
						MsgUndoBooking smsg=new MsgUndoBooking(rmsg.getFlyId(), rmsg.getSeat()[0], rmsg.getSeat()[1], ((MsgDeleteBooking) rmsg).getClientId(), InetAddress.getLocalHost().getHostAddress());
						//creazione processo di cancellazione ditribuita
						DistribuitedUndo regt=new DistribuitedUndo(smsg, this, HostConfig);
						Log.srvLogging(getName()+": wait distribuited request result....");
						regt.start();
						try {regt.join();}
						catch (InterruptedException e) {/*ignored*/}
						goon=true;
						//NB. la risposta (Rsp) e settata dal DeamonBook  o dal DistribuitedDo
						synchronized(database)
							{//cancellazione prenotazione locale 
							int pos_fly=database.posFlight(rmsg.getFlyId());
							if (pos_fly>=0)
								database.delBooking(pos_fly, rmsg.getSeat()[0], rmsg.getSeat()[1]);
							}
						//invio al cliente della risposta
						ClientWR.writeByte(Message.codeMsg("RSP"));
						ClientWR.writeObject(Rsp);
						Log.srvLogging(getName()+": send response");
						}
					else
						Log.srvLogging("($$$) - [DELETE_BOOKING] "+getName()+":  FLY NOT EXIST");	
					}
				catch (InterruptedException e) 		{Log.srvLogging("($$$) - [DELETE_BOOKING] "+getName()+" "+e);}
				catch (ClassNotFoundException e) 	{Log.srvLogging("($$$) - [DELETE_BOOKING] "+getName()+" "+e);}
				catch (Exception e)			{Log.srvLogging("($$$) - [DELETE_BOOKING] "+getName()+" "+e);}
				}
			if (!okmsgrcv)
			//altri messaggi
				try {	//messaggio ricostruito e ignorato
					Object ignored=ClientRD.readObject();
					Log.netLogging("($$$) "+getName()+":  - INCONSISTENT CLIENT MESSAGGE");}
				catch (ClassNotFoundException e) {Log.srvLogging(getName()+" "+e);}
			}//end while
			}//end try
		//scadenza time-out
		catch (InterruptedIOException e)	{
							Log.srvLogging("($$$) - [CLIENT DOWN] "+getName());
							Log.srvLogging(e.toString());
							try 	{
								ClientWR.close();
								ClientRD.close();
								ClientSock.close();
								}
							catch(IOException e1) {/*ignored*/}
							}
		catch (StreamCorruptedException e) 	{Log.srvLogging("($$$) - [SOCKET STREAM CORRUPTED] "+getName()+" "+e);}
		catch (EOFException e)			{
							Log.srvLogging("($$$) - [CLIENT LEAVE] "+getName());
							try 	{
								ClientWR.close();
								ClientRD.close();
								ClientSock.close();
								}
							catch(IOException e1) {/*ignored*/}
							} 
		catch (IOException e) 			{Log.srvLogging("($$$) - [EXCEPTION STREAM] "+getName()+" "+e);}
		}	
	}
